#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <pcre.h>

#include "pages.h"

char HTTP_SERVER[] = "UFO/2.3.1";

char HTTP_MSG[] = "HTTP/1.1 %d %s\n"
                  "Server: %s\n"
                  "Content-Type: text/html; charset=UTF-8\n"
                  "Content-Length: %d\n"
                  "Connection: close"
                  "\n\n"
                  "%s";

void http_msg(char *buf, int code, char* message, char* body){
    sprintf(buf, HTTP_MSG, code, message, HTTP_SERVER, strlen(body), body);
}

int http_parse_request(char *str, char* method, char* uri, char* proto, char *headers, char* body){
    int n = sscanf(str, "%20s %2048s %10s\n", method, uri, proto);
    int i;
    char *p = strstr(str, "\r\n");
    if(!p) return 0;
    p += 2;
    // headers
    for(i=0; p[i]!=0; i++){
        if(i >= 3 && p[i-3] == '\r' && p[i-2] == '\n'  && p[i-1] == '\r' && p[i] == '\n'){
            headers[i-3] = 0;
            break;
        }
        headers[i] = p[i];
    }
    headers[i] = 0;
    // body
    for(; p[i]!=0; i++) body[i] = p[i];
    body[i] = 0;
    return n;
}

int http_send_msg(int sock, char* client, int code, char* message, char* body){
    char buf[10240]; 
    http_msg(buf, code, message, body);
    int n = write(sock, buf, strlen(buf));
    if(n < 0){
        perror("ERROR writing to socket");
        return 1;
    }
    printf("[%s] %d %s\n", client, code, message);
    return 0;
}

int http_read_msg(int sock, char* method, char* uri, char* proto, char *headers, char* body){
    char buf[10240];
    int n = read(sock, buf, 10240);
    if (n < 0) {
        perror("ERROR reading from socket");
        return 1;
    }

    n = http_parse_request(buf, method, uri, proto, headers, body);
    return n < 3 ? 2 : 0;
}

int doprocessing(int sock, char *client) {
    char method[20], uri[1024], proto[10], headers[1024], body[2048];

    int n = http_read_msg(sock, method, uri, proto, headers, body);
    printf("method: %s\nuri: %s\nproto: %s\nheaders: {\n%s\n}\nbody: %s\n", method, uri, proto, headers, body);

    if(n > 0){
        http_send_msg(sock, client, 400, "Bad request", BAD_REQUEST_PAGE);
        return 2;
    }

    printf("[%s] %s %s %s\n", client, method, uri, proto);

    if(strcmp(method, "GET") != 0){
        http_send_msg(sock, client, 501, "Not Implemented", NOT_IMPLEMENTED_PAGE);
    }
    else if(strcmp(uri, "/") == 0){
        http_send_msg(sock, client, 200, "Ok", INDEX_PAGE);
    }
    else {
        http_send_msg(sock, client, 404, "Not found.", NOT_FOUND_PAGE);
    }

    return 0;
}

int main(int argc, char *argv[]) {
    int sockfd, newsockfd, portno, clilen;
    char buffer[256];
    struct sockaddr_in serv_addr, cli_addr;
    int n;

    if(argc < 2) {
        printf("Enter port number\n");
        return 1;
    }

    portno = atoi(argv[1]);

    /* First call to socket() function */
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) {
        perror("ERROR opening socket");
        return 1;
    }
    /* Initialize socket structure */
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = INADDR_ANY;
    serv_addr.sin_port = htons(portno);

    /* Now bind the host address using bind() call.*/
    if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
       perror("ERROR on binding");
       return 1;
    }
    /* Now start listening for the clients, here
     * process will go in sleep mode and will wait
     * for the incoming connection
    */
    listen(sockfd,5);
    clilen = sizeof(cli_addr);
    while (1) {
        newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
        if (newsockfd < 0) {
            perror("ERROR on accept");
            return 1;
        }
        /* Create child process */
        pid_t pid = fork();
        if (pid < 0) {
            perror("ERROR on fork");
            return 1;
        }
        if (pid == 0) {
            /* This is the client process */
            close(sockfd);
            char *client = inet_ntoa(cli_addr.sin_addr);
            int r = doprocessing(newsockfd, client);
            printf("[%s] done with code %d\n", client, r);
            return r;
        }
        else
        {
            close(newsockfd);
        }
    } /* end of while */
    return 0;
}
